
         /*  graftext.c,  Graphics Text Routines for Turbo C  */
         /*                                                   */
         /*          Copyright (c) 1990, John Ryon            */


#include  <graphics.h>


typedef unsigned char uchar;


static uchar FirstChar = 32,    /* First char to setup */
             LastChar = 126,    /* Last char to setup */
             CharHigh,          /* Current char height */
             CharWide[256];     /* Current char widths */


/*----------------------- SetCharRange ------------------------------------*/
/*                                                                         */
/*     Set the range of characters to include in the size table.           */
/*     The default range is the printable characters (isprint true)        */

void SetCharRange( uchar NewFirstChar, uchar NewLastChar )
{
   FirstChar = NewFirstChar;
   LastChar  = NewLastChar;
}


/*----------------------- SetupCharSizeTable ------------------------------*/
/*                                                                         */
/*           Setup a table of character sizes for the current              */
/*           font and charsize combination                                 */

static void SetupCharSizeTable( void )
{
   char Char[] = " ";  int C;

   CharHigh = textheight( "A" );
   for ( C = FirstChar;  C <= LastChar;  C++ )
   {
      Char[0] = C;  CharWide[C] = textwidth( Char );
   }
}


/*------------------------------ GCD --------------------------------------*/
/*                                                                         */
/*           Calculate the greatest common divisor of two integers.        */

static int GCD( int A, int B )
{
   int T;

   if ( A < B ) { T = A;  A = B;  B = T; }
   do {
      A = A % B;
      T = A;  A = B;  B = T;
   } while ( B != 0 );
   return A;
}

/*--------------------------- Reduce --------------------------------------*/
/*                                                                         */
/*       Reduce two integers by dividing both by their greatest            */
/*       common divisor                                                    */

static void Reduce( int *A, int *B )
{
   int D;

   D = GCD( *A, *B );
   *A /= D;
   *B /= D;
   while ( *A > getmaxx()  ||  *B > getmaxy() )
   {
      *A = (*A+1)/2;   *B = (*B+1)/2;
   }
}


static int CurrentFont = -1,     /* Current font */
           CurrentDir = -1,      /* Current direction (Hor | Ver) */
           CurrentSize = -1;     /* Current size number */

static int XM = 1,    /* Current XMul scale factor */
           XD = 1,    /* Current XDiv scale factor */
           YM = 1,    /* Current YMul scale factor */
           YD = 1;    /* Current YDiv scale factor */

/*------------------------ SetUserCharSize --------------------------------*/
/*                                                                         */
/*       Replacement for the BGI setusercharsize.                          */
/*       It generates a character size table when necessary                */
/*       (the call would change the usercharsize settings).                */

void SetUserCharSize( int NewXMul, int NewXDiv, int NewYMul, int NewYDiv )
{
   Reduce( &NewXMul, &NewXDiv );  Reduce( &NewYMul, &NewYDiv );

   if ( CurrentSize == 0  &&                   /* if no change, do nothing */
        NewXMul == XM  &&  NewXDiv == XD  &&
        NewYMul == YM  &&  NewYDiv == YD ) return;  

   XM = NewXMul;  XD = NewXDiv;  YM = NewYMul,  YD = NewYDiv;

   if ( CurrentSize != 0 )
   {
      setusercharsize( 1, 1, 1, 1 );
      SetupCharSizeTable();
   }
   setusercharsize( XM, XD, YM, YD );
   CurrentSize = 0;
}

/*-------------------------- TextWidth ------------------------------------*/
/*                                                                         */
/*       Replacement for the BGI textwidth.                                */
/*       This one does the job correctly.                                  */

int TextWidth( uchar *Text )
{
   int Wide = 0;
   if ( CurrentSize == 0 )
      while ( *Text != '\0' )  Wide += CharWide[ *Text++ ] * XM / XD;
   else
      while ( *Text != '\0' )  Wide += CharWide[ *Text++ ];
   return Wide;
}

/*------------------------ GetTextSize ------------------------------------*/
/*                                                                         */
/*       Computes textwidth and textheight together & correctly.           */

void GetTextSize( uchar *Text, int *Wide, int *High )
{
   if ( CurrentSize == 0 )
   {
      *High = CharHigh * YM / YD;
      *Wide = 0;
      while ( *Text != '\0' )  *Wide += CharWide[ *Text++ ] * XM / XD;
   }
   else
   {
      *High = CharHigh;
      *Wide = 0;
      while ( *Text != '\0' )  *Wide += CharWide[ *Text++ ];
   }
}

/*------------------------ GetStdTextSize ---------------------------------*/
/*                                                                         */
/*      Computes textwidth & textheight without using the                  */
/*      usercharsize scale factors.                                        */

void GetStdTextSize( uchar *Text, int *Wide, int *High )
{
   *High = CharHigh;
   *Wide = 0;
   while ( *Text != '\0' )  *Wide += CharWide[ *Text++ ];
}


/*------------------------ SetTextSize ------------------------------------*/
/*                                                                         */
/*     Sets the usercharsize scale factors to display the                  */
/*     text string with the given width and height (in pixels).            */
/*     If Wide is negative, the width scale is set equal to                */
/*     the scale used for the height.                                      */

void SetTextSize( char *Text, int Wide, int High )
{
   int OrigWide, OrigHigh;

   if ( High <= 0 ) return;
   if ( CurrentSize != 0 )
      SetUserCharSize( 1, 1, 1, 1 );
   GetStdTextSize( Text, &OrigWide, &OrigHigh );
   if ( Wide > 0 )
      SetUserCharSize( Wide, OrigWide, High, OrigHigh );
   else
      SetUserCharSize( High, OrigHigh, High, OrigHigh );
}

/*------------------------ SetEmSize --------------------------------------*/
/*                                                                         */
/*     Sets the usercharsize scale factors to display a square             */
/*     character such as the letter 'm' with the given width &            */
/*     height (in pixels).  Both parameters must be greater than zero.     */

void SetEmSize( int Wide, int High )
{
   int  OrigWide = CharHigh,  OrigHigh = CharHigh;

   if ( Wide <= 0  ||  High <= 0 ) return;
   SetUserCharSize( Wide, OrigWide, High, OrigHigh );
}


/*------------------------ SetTextStyle -----------------------------------*/
/*                                                                         */
/*     This is a replacement for the BGI settextstyle.                     */
/*     It generates a character size table when necessary                  */
/*     (font or charsize is changed)                                       */

void SetTextStyle( int NewFont, int NewDirection, int NewSize )
{
   if ( NewFont == CurrentFont  &&      /* if no change, do nothing */
        NewDirection == CurrentDir  &&  NewSize == CurrentSize ) return;

   settextstyle( NewFont, NewDirection, NewSize );
   if ( NewFont != CurrentFont  ||  NewSize != CurrentSize )
   {
      if ( NewSize == 0 )
      {
         setusercharsize( 1, 1, 1, 1 );
         SetupCharSizeTable();
         setusercharsize( XM, XD, YM, YD );
         settextstyle( NewFont, NewDirection, NewSize );
      }
      else SetupCharSizeTable();
   }

   CurrentFont = NewFont;
   CurrentDir = NewDirection;
   CurrentSize = NewSize;
}


static struct textsettingstype TS;

/*------------------------ JustifyText ------------------------------------*/
/*                                                                         */
/*     This uses the correct text size to justify the given                */
/*     text string about the given pixel position.                         */

static int JustifyText( int *I, int *J, char *Text )
{
   int Wide, High;

   GetTextSize( Text, &Wide, &High );
   switch ( TS.horiz )
   {
   case LEFT_TEXT:   break;
   case CENTER_TEXT: *I -= Wide / 2;    break;
   case RIGHT_TEXT:  *I -= (Wide-1);    break;
   }
   switch ( TS.vert )
   {
   case TOP_TEXT:   break;
   case CENTER_TEXT: *J -= High / 2;    break;
   case BOTTOM_TEXT:  *J -= (High-1);    break;
   }
   return Wide;
}

/*------------------------ OutTextXY --------------------------------------*/
/*                                                                         */
/*     This is a replacement for the BGI outtextxy.                             */
/*     It uses the correct text justification.                             */

void OutTextXY( int I, int J, char *Text )
{
   gettextsettings( &TS );
   JustifyText( &I, &J, Text );
   settextjustify( LEFT_TEXT, TOP_TEXT );
   outtextxy( I, J, Text );
   settextjustify( TS.horiz, TS.vert );
}

/*------------------------ OutText ----------------------------------------*/
/*                                                                         */
/*     This is a replacement for the BGI outtext.                               */
/*     It uses the correct text justification.                             */

void OutText( char *Text )
{
   int I, J, Wide;

   gettextsettings( &TS );  I = getx();  J = gety();
   Wide = JustifyText( &I, &J, Text );
   settextjustify( LEFT_TEXT, TOP_TEXT );
   outtextxy( I, J, Text );
   settextjustify( TS.horiz, TS.vert );
   if ( TS.horiz == LEFT_TEXT  &&  TS.direction == HORIZ_DIR )
      moverel( Wide, 0 );
}

/*------------------------ OutTextWH --------------------------------------*/
/*                                                                         */
/*     This is a combination of SetTextSize and OutText.                        */
/*     It uses the correct text justification.                             */

void OutTextWH( int Wide, int High, char *Text )
{
   SetTextSize( Text, Wide, High );
   OutText( Text );
}

/*------------------------ OutTextXYWH ------------------------------------*/
/*                                                                         */
/*     This is a combination of SetTextSize and OutTextXY>                      */
/*     It uses the correct text justification.                             */

void OutTextXYWH( int I, int J, int Wide, int High, char *Text )
{
   SetTextSize( Text, Wide, High );
   OutTextXY( I, J, Text );
}

